package tdspp.Weather;

import tdspp.util.*;
import java.util.*;
import java.io.*;
import Prism.core.*;

/**
 * Stores the weather information tagged to regions.
 *
 * This will store the weather inforation tagged to the regions which the system is composed of.
 * It does not need to store the regions objects. Thus it will have a collection of region
 * handles and corresponding weather information.
 * 
 * <table border=0> 
 * <tr> <td style="font-weight: bold; font-size: smaller"> Belongs to </td> <td> TDS++ </td>
 * <tr> <td style="font-weight: bold; font-size: smaller"> Properties </td> <td> Optional, Passive </td>
 * </table> 
 * 
 * @author 	TDS++ team, USC-CSE
 * @version 	1.0 , 17 Feb 2002
 * 
*/

public class WeatherImp extends GenericImp {
	/**
	 * Weather data : each element is itself an array of Region objects.
	 * Each element in turn has 10x10 objects arranged in row-major order.
	 */
	Object[] array;

	/**
	 * Gives the probability of a region weather update. The probabilities are all stored as in range 0 to 1.
	 */
	double updateProb;
	
	/**
	 * Probability of temperature update if a region's weather is being updated.
	 * A random number is generated if the number is less than half of the probability
	 * then a negative update of the tempIncr is done, if the number is greater than half of tempProb
	 * but less than tempProb then a positive update of tempIncr is done, else no change.
	 */
	double tempProb;
	
	/**
	 * Probability of humidity update if a region's weather is being updated.
	 */
	double humidProb;
	
	/**
	 * Probability of windSpeed update if a region's weather is being updated.
	 */
	double windProb;
	
	/**
	 * Probability of sunlight update if a region's weather is being updated.
	 */
	double sunlightProb;
	
	/**
	 * Change in temperature.
	 */
	double tempIncr;
	
	/**
	 * Change in humidity.
	 */
	double humidIncr;
	
	/**
	 * Change in windSpeed.
	 */
	double windIncr;
	
	/**
	 * Change in sunlight.
	 */
	double sunlightIncr;

	/**
	 * Constructor
	 * 
	*/
	public WeatherImp(AbstractImplementation parent) {
		super(parent);
		array = null;
	}

	/**
	 * Set the probabilities and increments
	 * 
	*/
	public void setUpdateParams(double updateProb, 
						double tempProb, double humidProb, double windProb,	double sunlightProb,
						double tempIncr, double humidIncr, double windIncr, double sunlightIncr) {
		this.updateProb = updateProb;
		this.tempProb = tempProb;
		this.humidProb = humidProb;
		this.windProb = windProb;
		this.sunlightProb = sunlightProb;
		this.tempIncr = tempIncr;
		this.humidIncr = humidIncr;
		this.windIncr = windIncr;
		this.sunlightIncr = sunlightIncr;
	}

	/**
	 * Provides the weather information for a particular region. Thus it checks for the region handle
	 * returns the corresponding weather object.
	 * Called by : WeatherAnalyzer
	 * 
	 * @param r	The region handle.
	 * @see 	WeatherAnalyzerImp#getWeather
	 * @return	WeatherInfo
	 * 
	*/
	public WeatherInfo getWeather(Region r) {
		Region[] regions = (Region[]) array[r.submapId];
		for (int i = 0; i < regions.length; i++) {
			Region reg = regions[i];
			if(r.x == reg.x && r.y == reg.y)
				return reg.weatherInfo;
		}		
		return null;
	}
	
	/**
	 * Populate the weatherInfo in the list of the Regions passed in.
	 * Called by : WeatherAnalyzer
	 * 
	 * @param regions Vector of Region objects
	 * 
	*/
	public void populateWeatherInfo(Region[] ipRegions) {
		for (int j = 0; j < ipRegions.length; j++) {
			Region ipreg = ipRegions[j];
			Region[] regions = (Region[]) array[ipreg.submapId];
			for (int i = 0; i < regions.length; i++) {
				Region reg = regions[i];
				if(ipreg.x == reg.x && ipreg.y == reg.y) {
					ipreg.weatherInfo = reg.weatherInfo;
					break;
				}
			}
		}		
	}

	/**
	 * Update the weather info.
	 * 
	*/
	public void update() {
		for (int i = 0; i < array.length ; i++) {
			Region[] regions = (Region[]) array[i];
			for (int j = 0; j < regions.length; j++) {
				Region reg = regions[j];
				// update the weather for this region
				double rndm = Math.random();
				if(rndm < updateProb) {
					// yes update
					rndm = Math.random();
					if(rndm < tempProb) {
						if(rndm < (tempProb/2))
							reg.weatherInfo.temperature -= tempIncr;
						else 
							reg.weatherInfo.temperature += tempIncr;
					}
					rndm = Math.random();
					if(rndm < humidProb) {
						if(rndm < (humidProb/2))
							reg.weatherInfo.humidity -= humidIncr;
						else 
							reg.weatherInfo.humidity += humidIncr;
					}
					rndm = Math.random();
					if(rndm < windProb) {
						if(rndm < (windProb/2))
							reg.weatherInfo.windSpeed -= windIncr;
						else 
							reg.weatherInfo.windSpeed += windIncr;
					}
					rndm = Math.random();
					if(rndm < sunlightProb) {
						if(rndm < (sunlightProb/2))
							reg.weatherInfo.sunlight -= sunlightIncr;
						else 
							reg.weatherInfo.sunlight += sunlightIncr;
					}
				}
			}
		}
	}
	
	/**
	 * Read the file and read the initial data
	 */
	public void readFile(String filename) {
		try {
			File file = new File(filename);
			BufferedReader br = new BufferedReader(new FileReader(file));
			String line;
			int lineNum = 0;
			if(array == null) {
				array = new Object[Constants.REPOSITORY_NUMBER_OF_SUBMAPS];
				for(int i = 0; i < Constants.REPOSITORY_NUMBER_OF_SUBMAPS; i++) {
					Region[] subarr = new Region[Constants.REPOSITORY_CELLS_PER_SUBMAP * Constants.REPOSITORY_CELLS_PER_SUBMAP];
					for(int j = 0; j < Constants.REPOSITORY_CELLS_PER_SUBMAP * Constants.REPOSITORY_CELLS_PER_SUBMAP; j++) {
						subarr[j] = new Region();
						subarr[j].submapId = i;
						subarr[j].x = j / Constants.REPOSITORY_CELLS_PER_SUBMAP;
						subarr[j].y = j % Constants.REPOSITORY_CELLS_PER_SUBMAP;
						subarr[j].weatherInfo = new WeatherInfo();
					}
					array[i] = subarr;
				}
			}
			while((line = br.readLine()) != null) {
				line = line.trim();
				if(line.charAt(0) == ';' || line.charAt(0) == ':') continue;
				if(lineNum < Constants.REPOSITORY_NUMBER_OF_SUBMAPS) {
					// reading image names
					// System.out.println("Weather.readFile : Image " + line);
					lineNum++;
				} else {
					// reading Region data
					//System.out.println("Weather.readFile : Region " + line);
					// cx cy submapid temperature humidity windSpeed sunlight resources stateAmmu stateFood stateMed terraintype
					StringTokenizer st = new StringTokenizer(line);
					int tokens = st.countTokens();
					if(tokens > 3) {
						// atleast 4 tokens required
						Region r = new Region();
						r.x = Integer.parseInt(st.nextToken());
						tokens--;
						r.y = Integer.parseInt(st.nextToken());
						tokens--;
						r.submapId = Integer.parseInt(st.nextToken());
						tokens--;
						r.weatherInfo = new WeatherInfo();
						r.weatherInfo.temperature = new Float(st.nextToken()).floatValue();
						tokens--;
						if(tokens-- > 0) r.weatherInfo.humidity = new Float(st.nextToken()).floatValue();
						if(tokens-- > 0) r.weatherInfo.windSpeed = new Float(st.nextToken()).floatValue();
						if(tokens-- > 0) r.weatherInfo.sunlight = Integer.parseInt(st.nextToken());
						setWeather(r);
					}
				}
			}
			br.close();
		} catch(Exception e) {
			System.out.println("Weather.readFile : " + e);
		}
	}

	/**
	 * Sets the weather info
	 * 
	 * @param r	Region whose attributes are to be changed.
	 * @return	true or false
	 * 
	*/
	public boolean setWeather(Region r) {
		Region[] regions = (Region[]) array[r.submapId];
		for (int i = 0; i < regions.length ; i++) {
			Region reg = regions[i];
			if(r.x == reg.x && r.y == reg.y) {
				regions[i] = r;
				return true;
			}
		}
		return false;
	}

}
